﻿library wrwer;

{ For 64-bit Delphi XE2: Project options - Build Events - select platforms - add to
  'Post Build Events - Commands' line "copy diskdir.dll diskdirW.wcx64" (without quotes) for 64-bit
  and "copy diskdir.dll diskdirW.wcx" for 32-bit plugin.
  For Delphi XE3 and up just use $E directive }

{$IFDEF FPC}
  {$MODE Delphi}
  {$ASMMODE INTEL}
{$ELSE}
  {$WEAKLINKRTTI ON}
{$ENDIF}

{$R diskdir.res}

uses
  SysUtils,Classes,Windows,IniFiles;

type
  char=ansichar;
  pchar=pansichar;

const       {Error codes returned to calling application}
  E_END_ARCHIVE=     10;       {No more files in archive}
  E_NO_MEMORY=       11;       {Not enough memory}
  E_BAD_DATA=        12;       {Data is bad}
  E_BAD_ARCHIVE=     13;       {CRC error in archive data}
  E_UNKNOWN_FORMAT=  14;       {Archive format unknown}
  E_EOPEN=           15;       {Cannot open existing file}
  E_ECREATE=         16;       {Cannot create file}
  E_ECLOSE=          17;       {Error closing file}
  E_EREAD=           18;       {Error reading from file}
  E_EWRITE=          19;       {Error writing to file}
  E_SMALL_BUF=       20;       {Buffer too small}
  E_EABORTED=        21;       {Function aborted by user}
  E_NO_FILES=        22;       {No files found}
  E_TOO_MANY_FILES=  23;       {Too many files to pack}
  E_NOT_SUPPORTED=   24;       {Function not supported}

  {Unpacking flags}
  PK_OM_LIST=           0;
  PK_OM_EXTRACT=        1;

  {Flags for ProcessFile}
  PK_SKIP=              0;     {Skip file (no unpacking)}
  PK_TEST=              1;     {Test file integrity}
  PK_EXTRACT=           2;     {Extract file to disk}

  {Flags passed through ChangeVolProc}
  PK_VOL_ASK=           0;     {Ask user for location of next volume}
  PK_VOL_NOTIFY=        1;     {Notify app that next volume will be unpacked}

  {Packing flags}

  {For PackFiles}
  PK_PACK_MOVE_FILES=   1;    {Delete original after packing}
  PK_PACK_SAVE_PATHS=   2;    {Save path names of files}

  {Returned by GetPackCaps}
  PK_CAPS_NEW=          1;    {Can create new archives}
  PK_CAPS_MODIFY=       2;    {Can modify exisiting archives}
  PK_CAPS_MULTIPLE=     4;    {Archive can contain multiple files}
  PK_CAPS_DELETE=       8;    {Can delete files}
  PK_CAPS_OPTIONS=     16;    {Supports the options dialogbox}

type
//  pbyte=^byte;
  {Definition of callback functions called by the DLL}
  {Ask to swap disk for multi-volume archive}
  PChangeVolProc=^TChangeVolProc;
  TChangeVolProc=function(ArcName:pchar;Mode:longint):longint; stdcall;
  PChangeVolProcW=^TChangeVolProcW;
  TChangeVolProcW=function(ArcName:pwidechar;Mode:longint):longint; stdcall;
  {Notify that data is processed - used for progress dialog}
  PProcessDataProc=^TProcessDataProc;
  TProcessDataProc=function(Addr:pbyte;Size:longint):longint; stdcall;
  PProcessDataProcW=^TProcessDataProcW;
  TProcessDataProcW=function(FileName:pwidechar;Size:longint):longint; stdcall;

type
  THeaderData=packed record
    ArcName:array [0..259] of ansichar;
    FileName:array [0..259] of ansichar;
    Flags,
    PackSize,
    UnpSize,
    HostOS,
    FileCRC,
    FileTime,
    UnpVer,
    Method,
    FileAttr:longint;
    CmtBuf:pchar;
    CmtBufSize,
    CmtSize,
    CmtState:longint;
  end;

  THeaderDataEx=packed record
    ArcName:array [0..1023] of char;
    FileName:array [0..1023] of char;
    Flags,
    PackSize,
    PackSizeHigh,
    UnpSize,
    UnpSizeHigh,
    HostOS,
    FileCRC,
    FileTime,
    UnpVer,
    Method,
    FileAttr:longint;
    CmtBuf:pchar;
    CmtBufSize,
    CmtSize,
    CmtState:longint;
    Reserved:array[0..1023] of char;
  end;

  THeaderDataExW=packed record
    ArcNameW:array [0..1023] of widechar;
    FileNameW:array [0..1023] of widechar;
    Flags,
    PackSize,
    PackSizeHigh,
    UnpSize,
    UnpSizeHigh,
    HostOS,
    FileCRC,
    FileTime,
    UnpVer,
    Method,
    FileAttr:longint;
    CmtBuf:pwidechar;
    CmtBufSize,
    CmtSize,
    CmtState:longint;
    Reserved:array[0..1023] of AnsiChar;
  end;

  tOpenArchiveData=packed record
    ArcName:pchar;
    OpenMode,
    OpenResult:longint;
    CmtBuf:pchar;
    CmtBufSize,
    CmtSize,
    CmtState:longint;
  end;

  tOpenArchiveDataW=packed record
    ArcName:pwidechar;
    OpenMode,
    OpenResult:longint;
    CmtBuf:pwidechar;
    CmtBufSize,
    CmtSize,
    CmtState:longint;
  end;

const maxarchives=10;
type ArchiveRec=record
       ArchiveName:array[0..1023] of widechar;
       SrcDir,LastCurDir:array[0..1023] of widechar;
       CurrentFileName:array[0..1023] of widechar;
       CurrentUnpSize:comp;
       CurrentFileTime:longint;
       ArchiveHandle:textfile;
       ArchiveMode:longint;
       ArchiveActive:boolean;
       ChangeVolProc:TChangeVolProcW;
       ProcessDataProc:TProcessDataProcW;
     end;

var
  DontConvertCurrentLocaleToUTF8: Boolean = False;
  AddBOM: Boolean = False;
  UTF8Only: Boolean = False;

{Keep a list of currently open archives (up to a maximum of maxarchives)}
var ArchiveList:array[1..maxarchives] of ArchiveRec;

var _TokenSource: pwidechar;

FUNCTION StrTok(Source: pwidechar; Token: widechar): pwidechar; inline;
  VAR P: pwidechar;
BEGIN
  IF Source <> Nil THEN _TokenSource := Source;
  IF _TokenSource = Nil THEN begin
    strTok:=nil;
    exit
  end;
  P := StrScan(_TokenSource, Token);
  StrTok := _TokenSource;
  IF P <> Nil THEN BEGIN
    P^ := #0;
    Inc(P);
  END;
  _TokenSource := P;
END;

function GetVeryLongFileName(const FullFileName: string): string; inline;
begin
  Result := FullFileName;
  if Pos('\\', FullFileName) > 0 then
    Result := '\\?\UNC\' + Copy(FullFileName, 3, Length(FullFileName) - 2)
  else
    Result := '\\?\' + FullFileName;
end;

const bytesFromUTF8:array[AnsiChar] of byte = (
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 32
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 64
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 96
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  //128
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  //160
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  //192
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,  //224
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5); //256

function GetUtf8CharWidth(firstchar:Ansichar):integer;
begin
  result:=bytesFromUTF8[firstchar]+1;
end;

function IsFirstUTF8Char(thechar:Ansichar):boolean;
{The remaining bytes in a multi-byte sequence have 10 as their two most significant bits.}
begin
  result:=(byte(thechar) and (128+64))<>128;
end;

function IsSecondaryUTF8Char(thechar:Ansichar):boolean;
{The remaining bytes in a multi-byte sequence have 10 as their two most significant bits.}
begin
  result:=(byte(thechar) and (128+64))=128;
end;

function IsBufferUtf8(buf:pAnsichar;PartialAllowed:boolean):boolean;
{Buffer contains only valid UTF-8 characters, no secondary alone, no primary without the correct nr of secondary}
var p:pAnsichar;
    utf8bytes:integer;
    hadutf8bytes:boolean;
begin
  p:=buf;
  hadutf8bytes:=false;
  result:=false;
  utf8bytes:=0;
  while p[0]<>#0 do begin
    if utf8bytes>0 then begin  {Expecting secondary char}
      hadutf8bytes:=true;
      if not IsSecondaryUTF8Char(p[0]) then exit;  {Fail!}
      dec(utf8bytes);
    end else if IsFirstUTF8Char(p[0]) then
      utf8bytes:=GetUtf8CharWidth(p[0])-1
    else if IsSecondaryUTF8Char(p[0]) then
      exit;  {Fail!}
    inc(p);
  end;
    result:=hadutf8bytes and (PartialAllowed or (utf8bytes=0));
end;

function GetShortFileName(const sLongName: string): string; inline;
const
  cMaxPath = MAX_PATH + SizeOf(Char);
  cMaxPathWithTerm = cMaxPath - SizeOf(Char);
var
  sShortName:    string;
  nShortNameLen: Integer;
begin
  Result := '';
  SetLength(sShortName, cMaxPath);
  nShortNameLen := GetShortPathName(PWideChar(sLongName), PWideChar(sShortName), cMaxPathWithTerm);
  if (nShortNameLen = 0) or (nShortNameLen > cMaxPathWithTerm) then
    Exit;
  SetLength(sShortName, nShortNameLen);
  Result := sShortName;
end;

function GetShortFileNameEx(const sLongName: string): string;
begin
  Result := sLongName;
  if Length(Result) <= 256 then Exit;
  if Pos('\\?\', Result) = 0 then
  begin
    if Pos('\\', Result) > 0 then
      Result := '\\?\UNC\' + Copy(Result, 3, Length(Result) - 2)
    else
      Result := '\\?\' + Result;
  end;
  if Pos('\UNC\', Result) = 0 then
    Result := StringReplace(GetShortFileName(Result), '\\?\', '', []);
  if Result = '' then
    Result := sLongName;
end;

function ParseList(const Parsed: TStringList; ParseChar: PWideChar): Boolean;
var
  A: PWideChar;
begin
  Result := False;
  if not Assigned(Parsed) then Exit;
  try
    Parsed.Clear;
    if ParseChar = nil then Exit;
    A := ParseChar;
    while (A^ <> #0) do
    begin
      Parsed.Add(string(A));
      Inc(A, StrLen(A)+1);
    end;
    Result := True;
  except
    Result := False;
  end;
end;

function OpenArchiveW(var ArchiveData: tOpenArchiveDataW): Integer; stdcall;
var i:integer;
    archandle:integer;
    foundslot:boolean;
    BOM: array [0..2] of AnsiChar;
begin
  result:=0;
  foundslot:=false;
  for i:=1 to maxarchives do begin
    if not ArchiveList[i].ArchiveActive then begin
      foundslot:=true;
      archandle:=i;
      break;
    end;
  end;
  if not foundslot then begin
    ArchiveData.OpenResult:=E_NO_MEMORY;
    exit
  end;
  with ArchiveList[archandle] do begin
    LastCurDir[0]:=#0;
    strlcopy(ArchiveName, PWideChar(GetShortFileNameEx(ArchiveData.arcName)), Length(ArchiveName)-1);
    {$i-}
    if UTF8Only then
      assignfile(ArchiveHandle,ArchiveName, 65001)
    else
      assignfile(ArchiveHandle,ArchiveName);

    filemode:=0;
    reset(ArchiveHandle);

    { Check BOM }
    FillChar(BOM, SizeOf(BOM), 0);
    for i := Low(BOM) to High(BOM) do
      if not Eof(ArchiveHandle) then
        Read(ArchiveHandle, BOM[I])
      else
        Break;
    if (BOM[0] <> #$EF) or (BOM[1] <> #$BB) or (BOM[2] <> #$BF) then
      TTextRec(ArchiveHandle).BufPos := 0;

    {$i+}
    if ioresult<>0 then
      ArchiveData.OpenResult:=E_EOPEN
    else begin
      ArchiveActive:=true;
      SrcDir[0]:=#0;
      result:=archandle;
    end;
  end;
end;

function OpenArchive(var ArchiveData:tOpenArchiveData):thandle; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function CloseArchive(hArcData:thandle):longint; stdcall;
begin
  with ArchiveList[hArcData] do begin
    if ArchiveActive then begin
      {$i-}
      Close(ArchiveHandle);
      {$i+}
      if ioresult<>0 then
        Result:=E_ECLOSE
      else
        Result:=0;
      ArchiveActive:=false;
    end;
  end;
end;

type
  TDateTime1 = packed record
    Year, Month, Day, Hour, Min, Sec: Word;
  end;

procedure PackTime(var T: TDateTime1; var P: Longint); assembler; {$ifdef win32} pascal;{$endif}
asm
{$ifdef win32}
        PUSH EDI
        PUSH ESI
        PUSH EBX
        MOV EDI,[P]
        MOV  ESI,[T]
{$else}
        PUSH RDI
        PUSH RSI
        PUSH RBX
        MOV RDI,P
        MOV  RSI,T
{$endif}
	CLD
	LODSW
	SUB	AX,1980
	MOV	CL,9
	SHL	AX,CL
	XCHG	AX,DX
	LODSW
	MOV	CL,5
	SHL	AX,CL
	ADD	DX,AX
	LODSW
	ADD	DX,AX
	LODSW
	MOV	CL,11
	SHL	AX,CL
	XCHG	AX,BX
	LODSW
	MOV	CL,5
	SHL	AX,CL
	ADD	BX,AX
	LODSW
	SHR	AX,1
	ADD	AX,BX
	STOSW
	XCHG	AX,DX
	STOSW
        {$ifdef win32}
        POP EBX
        POP ESI
        POP EDI
        {$else}
        POP RBX
        POP RSI
        POP RDI
        {$endif}
end;

procedure UnpackTime(P: Longint; var T: TDateTime1); assembler; pascal;
asm
{$ifdef win32}
  PUSH    EBX
  PUSH    EDI
  MOV     EDI,T
{$else}
  PUSH    RBX
  PUSH    RDI
  MOV     RDI,T
{$endif}
  MOV EBX,P
	CLD
	MOV	EAX,EBX
	MOV	CL,9+16
	SHR	EAX,CL
	ADD	AX,1980
	STOSW
	MOV	EAX,EBX
	MOV	CL,5+16
	SHR	EAX,CL
	AND	AX,15
	STOSW
	MOV	EAX,EBX
	MOV	CL,16
	SHR	EAX,CL
	AND	AX,31
	STOSW
	MOV	AX,BX
	MOV	CL,11
	SHR	AX,CL
	STOSW
	MOV	AX,BX
	MOV	CL,5
	SHR	AX,CL
	AND	AX,63
	STOSW
	MOV	AX,BX
	AND	AX,31
	SHL	AX,1
	STOSW
{$ifdef win32}
        POP    EDI
        POP    EBX
{$else}
        POP    RDI
        POP    RBX
{$endif}
end;

function ReadHeader(hArcData:thandle;var HeaderData:THeaderData):longint; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function makecomp(lowlong,highlong:longint):comp; inline;
type tcomprec=record
       case boolean of
         true:(l,h:longint);
         false:(cmp:comp);
       end;
var tcr:tcomprec;
begin
  tcr.l:=lowlong;
  tcr.h:=highlong;
  result:=tcr.cmp;
end;

type tcomprec=record
       case boolean of
         true:(l,h:longint);
         false:(cmp:comp);
       end;

function ReadHeaderExW(hArcData: THandle; var HeaderData: tHeaderDataExW): Integer; stdcall;
var buf:array[0..2048] of widechar;
    buf1:array[0..2560] of widechar;
    p,p1,psize,pdate,ptime,pattr: pwidechar;
    tdt:TDateTime1;
    code:integer;
    UnpSizeComp:comp;
    ABuf: AnsiString;
    ABufW: WideString;

  procedure CheckUTF8;
  begin
    if IsBufferUtf8(PAnsiChar(ABuf), True) then
      ABufW := UTF8ToWideString(ABuf)
    else
      ABufW := ABuf;
  end;

begin
  if ArchiveList[hArcData].ArchiveActive then begin
    if not ArchiveList[hArcData].ArchiveActive then
      Result:=E_BAD_ARCHIVE
    else if eof(ArchiveList[hArcData].ArchiveHandle) then
      Result:=E_END_ARCHIVE
    else begin
      {$i-}
      if UTF8Only then
        readln(ArchiveList[hArcData].ArchiveHandle, ABufW)
      else begin
        readln(ArchiveList[hArcData].ArchiveHandle, ABuf);
        CheckUTF8;
      end;

      strlcopy(buf, PWideChar(ABufW), SizeOf(buf) - 1);

      if ArchiveList[hArcData].SrcDir[0]=#0 then begin   {First line!}
        if strscan(buf,#9)=nil then begin
          strlcopy(ArchiveList[hArcData].SrcDir, buf, SizeOf(ArchiveList[hArcData].SrcDir) - 1);

          if UTF8Only then
            readln(ArchiveList[hArcData].ArchiveHandle, ABufW)
          else begin
            readln(ArchiveList[hArcData].ArchiveHandle, ABuf);
            CheckUTF8;
          end;

          strlcopy(buf, PWideChar(ABufW), SizeOf(buf) - 1);
        end;
      end;
      {$i+}
      if ioresult=0 then begin
        fillchar(HeaderData,sizeof(HeaderData)-16,#0);
        p:=strtok(buf,#9);
        psize:=strtok(nil,#9);
        pdate:=strtok(nil,#9);
        ptime:=strtok(nil,#9);
        pattr:=strtok(nil,#9);
        if (buf[0]<>#0) and (buf[strlen(buf)-1]='\') then
          HeaderData.FileAttr:=fadirectory;
        if strscan(buf,'\')=nil then begin  {No directory -> assume last given dir!}
          strlcopy(buf1,ArchiveList[hArcData].LastCurDir,sizeof(HeaderData.FileNameW)-1);
        end else begin
          if HeaderData.FileAttr=fadirectory then
            strlcopy(ArchiveList[hArcData].LastCurDir,buf,sizeof(ArchiveList[hArcData].lastcurdir)-1);
          buf1[0]:=#0;
        end;
        strlcat(buf1,buf,sizeof(buf1)-1);
        p:=strscan(buf1,':');
        if p<>nil then inc(p,2)
                  else p:=buf1;
        strlcopy(HeaderData.FileNameW,p,sizeof(HeaderData.FileNameW));
        if ArchiveList[hArcData].SrcDir[0]=#0 then   {First line!}
          strlcopy(ArchiveList[hArcData].SrcDir,buf1,3);

        strlcopy(HeaderData.ArcNameW,ArchiveList[hArcData].ArchiveName,sizeof(HeaderData.ArcNameW)-1);

        if psize<>nil then begin
          val(psize,UnpSizeComp,code);
          if code<>0 then begin
            HeaderData.UnpSize:=-1;
            HeaderData.UnpSizeHigh:=-1;
          end else begin
            HeaderData.UnpSize:=tcomprec(UnpSizeComp).l;
            HeaderData.UnpSizeHigh:=tcomprec(UnpSizeComp).h;
          end;
          HeaderData.PackSize:=HeaderData.UnpSize;
        end else begin
          HeaderData.UnpSize:=-1;
          HeaderData.UnpSizeHigh:=-1;
        end;
        HeaderData.PackSizeHigh:=HeaderData.UnpSizeHigh;
        HeaderData.FileTime:=-1;
        if pdate<>nil then begin {Year.month.day}
          fillchar(tdt,sizeof(tdt),#0);
          p1:=strtok(pdate,'.');
          val(p1,tdt.year,code);
          p1:=strtok(nil,'.');
          if (code=0) and (p1<>nil) then begin
            if tdt.year<1900 then if tdt.year<80 then inc(tdt.year,2000)
                                                 else inc(tdt.year,1900);
            val(p1,tdt.month,code);
            p1:=strtok(nil,'.');
            if (code=0) and (p1<>nil) then begin
              val(p1,tdt.day,code);
              if code=0 then begin
                if ptime<>nil then begin {hour:min:seconds}
                  p1:=strtok(ptime,':');
                  val(p1,tdt.hour,code);
                  p1:=strtok(nil,'.');
                  if (code=0) and (p1<>nil) then begin
                    val(p1,tdt.min,code);
                    p1:=strtok(nil,'.');
                    if (code=0) then
                      if (p1<>nil) then val(p1,tdt.sec,code);
                  end;
                end;
                PackTime(tdt,HeaderData.FileTime);
                { Same as PackTime
                HeaderData.FileTime := (tdt.Year - 1980) shl 25 or
                  tdt.Month shl 21 or tdt.Day shl 16 or tdt.Hour shl 11 or
                  tdt.Min shl 5 or tdt.Sec div 2;
                }
              end;
            end;
          end;
        end;
        if pattr <> nil then
        begin
          if StrScan(pattr, 'a') <> nil then
            HeaderData.FileAttr := HeaderData.FileAttr or faArchive;
          if StrScan(pattr, 'h') <> nil then
            HeaderData.FileAttr := HeaderData.FileAttr or faHidden;
          if StrScan(pattr, 'r') <> nil then
            HeaderData.FileAttr := HeaderData.FileAttr or faReadOnly;
          if StrScan(pattr, 's') <> nil then
            HeaderData.FileAttr := HeaderData.FileAttr or faSysFile;
        end;

        strlcopy(ArchiveList[hArcData].CurrentFileName,HeaderData.FileNameW,sizeof(ArchiveList[hArcData].CurrentFileName)-1);
        ArchiveList[hArcData].CurrentFileTime:=HeaderData.FileTime;
        ArchiveList[hArcData].CurrentUnpSize:=makecomp(HeaderData.UnpSize,
          HeaderData.UnpSizeHigh);
        Result:=0;
      end else
        Result:=E_EREAD;
    end;
  end;
end;

function createvalidoutfile(var g:file;trgfile:pwidechar;isadir:boolean):integer;
var err:integer;
    buf,buf0:array[0..1023] of widechar;
    p,p1:pwidechar;
begin
  {$i-}
  if not isadir then begin
    assign(g,trgfile);
    rewrite(g,1);
    err:=ioresult;
  end;
  strlcopy(buf,trgfile,sizeof(buf)-1);
  if (err=3) or isadir then begin  {path not found}
    p1:=strrscan(buf,'\');
    if p1<>nil then inc(p1);  {pointer to filename}
    p:=strscan(buf,'\');
    if p<>nil then p:=strscan(p+1,'\');
    while (p<>nil) and (p<p1) do begin
      p[0]:=#0;
      mkdir(strpas(buf));
      err:=ioresult;
      p[0]:='\';
      p:=strscan(p+1,'\');
    end;
    if isadir then begin
      result:=0;    {A directory -> ok}
      exit;
    end;
    rewrite(g,1);
    err:=ioresult;
  end;
  result:=err;
end;

function ProcessFileW(hArcData: THandle; Operation: Integer; DestPath, DestName: PWideChar): Integer; stdcall;
var srcfile,trgfile:array[0..1023] of widechar;
    f,g:file;
    buf:array[0..32767] of ansichar;
    p:pwidechar;
    dataread,datawritten,err:integer;
    filetime:tfiletime;
    showprogress:boolean;
    LongFileName: WideString;
    FS: TFileStream;
    AIOResult: Integer;
begin
  FS := nil;
  if (Operation=PK_EXTRACT) or (Operation=PK_TEST) then
  with ArchiveList[hArcData] do if ArchiveActive then try

    LongFileName := '';
    strcopy(srcfile,srcdir);
    strlcat(srcfile,CurrentFilename,sizeof(srcfile)-1);
    filemode:=0;
    repeat
      if StrLen(srcfile) > 255 then
      begin
        LongFileName := GetVeryLongFileName(srcfile);
        if not FileExists(LongFileName) then
          err := 1
        else begin
          err := 0;
          try
            FS := TFileStream.Create(LongFileName, fmOpenRead or fmShareDenyWrite);
          except
            err := 1;
          end;
        end;
      end
      else
      begin
        assign(f,srcfile);
        {$i-}
        reset(f,1);
        {$i+}
        err:=ioresult;
      end;
      if err>0 then
        if (@ChangeVolProc=nil) or
          (ChangeVolProc(srcfile,0)=0) then err:=-1
        else begin           {Remeber path given by user!}
          p:=AnsiStrPos(srcfile,CurrentFilename);
          if p<>nil then
            strlcopy(srcdir,srcfile,p-@srcfile);
        end;
    until err<=0;

    if err<>0 then begin result:=E_EOPEN; exit; end;

    showprogress:=(CurrentUnpSize>0) and (@ProcessDataProc<>nil);
    if destpath<>nil then begin
      strcopy(trgfile,destpath);
      if (trgfile[0]<>#0) and (trgfile[strlen(trgfile)-1]<>'\') then
        strcat(trgfile,'\');
    end else trgfile[0]:=#0;
    if (Operation=PK_EXTRACT) then begin
      if destname<>nil then
        strlcat(trgfile,destname,sizeof(trgfile)-1)
      else begin
        p:=strrscan(CurrentFileName,'\');
        if p<>nil then inc(p) else p:=CurrentFileName;
        strlcat(trgfile,CurrentFileName,sizeof(trgfile)-1);
      end;
      if createvalidoutfile(g,trgfile,false)<>0 then begin close(f); result:=E_ECREATE; exit; end;
    end;

    while true do begin
      {$i-}
      AIOResult := 0;
      if Length(LongFileName) > 0 then
      begin
        try
          dataread := FS.Read(buf[0],sizeof(buf));
          if dataread = 0 then
            AIOResult := 1
          else
            AIOResult := 0;
          if (dataread = 0) and (FS.Position = FS.Size) then
            AIOResult := 0;
          except
            AIOResult := 1
          end;
      end else begin
        blockread(f,buf[0],sizeof(buf),dataread);
        AIOResult := ioresult;
      end;

      if AIOResult<>0 then begin
        if (Operation=PK_EXTRACT) then close(g);
        result:=E_EREAD;
        exit
      end else begin
        if (Operation=PK_EXTRACT) then begin
          blockwrite(g,buf[0],dataread,datawritten);
          if AIOResult<>0 then begin
            close(g);
            result:=E_EWRITE;
            exit
          end
        end else
          datawritten:=dataread;
        if showprogress then begin
          if ProcessDataProc(srcfile,datawritten)=0 then begin {Abort!}
            if Length(LongFileName) = 0 then
              close(f);
            if (Operation=PK_EXTRACT) then close(g);
            result:=0;     {UnRAR.DLL returns no error when aborted!}
            exit;
          end;
        end;
      end;
      if dataread<sizeof(buf) then break;
      {$i+}
    end;
    if Length(LongFileName) = 0 then
      close(f);
    if (Operation=PK_EXTRACT) then begin
      FileSetDate(tfilerec(g).handle,CurrentFileTime);
      close(g);
    end;
  finally
    FS.Free;
  end;
  Result:=0;
end;

function ProcessFile(hArcData:thandle;Operation:longint;DestPath,DestName:pchar):longint; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

var PackerChangeVolProc:TChangeVolProcW;
    PackerProcessDataProc:TProcessDataProcW;

procedure SetChangeVolProcW(hArcData:THandle;ChangeVolProc1:TChangeVolProcW); stdcall;
begin
  if NativeInt(hArcData)=-1 then
    PackerChangeVolProc:=ChangeVolProc1
  else
    ArchiveList[hArcData].ChangeVolProc:=ChangeVolProc1;
end;

procedure SetChangeVolProc(hArcData:Integer;ChangeVolProc1:TChangeVolProc); stdcall;
begin
end;

procedure SetProcessDataProcW(hArcData:THandle;ProcessDataProc1:TProcessDataProcW); stdcall;
begin
  if NativeInt(hArcData)=-1 then
    PackerProcessDataProc:=ProcessDataProc1
  else
    ArchiveList[hArcData].ProcessDataProc:=ProcessDataProc1;
end;

procedure SetProcessDataProc(hArcData:Integer;ProcessDataProc1:TProcessDataProc); stdcall;
begin
end;

function DoUTF8Encode(const Str: WideString): AnsiString; inline;
begin
  if DontConvertCurrentLocaleToUTF8 then
  begin
    if AnsiString(Str) = Str then
      Result := AnsiString(Str)
    else
      Result := UTF8Encode(Str);
  end else
    Result := UTF8Encode(Str);
end;

function MakeAttrString(const [Ref] SR: TSearchRec): AnsiString; inline;
begin
  Result := '';
  if SR.Attr and faReadOnly <> 0 then
    Result := Result + 'r';
  if SR.Attr and faArchive <> 0 then
    Result := Result + 'a';
  if SR.Attr and faHidden <> 0 then
    Result := Result + 'h';
  if SR.Attr and faSysFile <> 0 then
    Result := Result + 's';
end;

function LeftPad(Value: Cardinal): AnsiString; inline;
begin
  if Value > 9 then
    Result := IntToStr(Value)
  else
    Result := '0' + IntToStr(Value)
end;

procedure ConvertTime(const FindData: _WIN32_FIND_DATA; var T: TDateTime1); inline;
var
  FT: TFileTime;
  ST: TSystemTime;
begin
  T := Default(TDateTime1);
  FileTimeToLocalFileTime(FindData.ftLastWriteTime, FT);
  if FileTimeToSystemTime(FT, ST) then
    with ST do
    begin
      T.Year := wYear;
      T.Month := wMonth;
      T.Day := wDay;
      T.Hour := wHour;
      T.Min := wMinute;
      T.Sec := wSecond;
    end;
end;

function PackFilesW(PackedFile, SubPath, SrcPath, AddList: PWideChar; Flags: Integer): Integer; stdcall;
var s:tsearchrec;
    PackedHandle:text;
    p,pEnd,pData:pwidechar;
    tdt:tdatetime1;
    buf:array[0..1299] of widechar;
    RunOk:boolean;
    ATmp: WideString;
    Atmp1: array[0..1400] of AnsiChar; // Extent when add new tab-separated "fields"
    ATMP2, Attr: AnsiString;
    AFindResult: Integer;
    SL: TStringList;
begin
  with TIniFile.Create(ChangeFileExt(GetModuleName(HInstance), '.ini')) do
  try
    DontConvertCurrentLocaleToUTF8 := ReadBool('diskdirW',
      'DontConvertCurrentLocaleToUTF8', False);
    AddBOM := ReadBool('diskdirW', 'AddBOM', False);
  finally
    Free;
  end;
  if UTF8Only then
    DontConvertCurrentLocaleToUTF8 := False;

  if FileExists(PackedFile) then begin
    {
    SL := TStringList.Create;
    try
      ParseList(SL, AddList);
      MessageBox(GetActiveWindow, PWideChar(SubPath +#13+ SrcPath +#13#13+ SL.Text), nil, 0);
    finally
      SL.Free;
    end;
    }
    result:=E_NOT_SUPPORTED;
    exit
  end else begin
    {$i-}
    assignfile(PackedHandle, PackedFile);
    filemode:=0;
    rewrite(PackedHandle);
    {$i+}
    if ioresult<>0 then
    begin
      Result:=E_EWRITE;
      exit
    end else begin
      if AddBOM and (not DontConvertCurrentLocaleToUTF8) then
      begin
        write(PackedHandle, AnsiChar($EF), AnsiChar($BB), AnsiChar($BF));
        //const UTF8: array [0..2] of byte = ($EF, $BB, $BF);
        //Move(PByte(@UTF8[0])^, PByte(TTextRec(PackedHandle).BufPtr)^, Length(UTF8));
        //TTextRec(PackedHandle).BufPos := 3;
      end;

      if SrcPath[0]<>#0 then            {Search result?}
      begin
        ATMP2 := AnsiString(DoUTF8Encode(SrcPath));
        StrLCopy(Atmp1, PAnsiChar(ATMP2), SizeOf(Atmp1) - 1);
        writeln(PackedHandle, Atmp1);
      end;
      p:=AddList;
      strlcopy(buf,SrcPath, SizeOf(buf) - 1);
      pEnd:=Strend(buf);
      RunOK:=true;
      while (p[0]<>#0) and RunOK do begin
        strlcat(buf,p,sizeof(buf)-1);
        if buf[strlen(buf)-1]='\' then
          buf[strlen(buf)-1]:=#0;
        if StrLen(buf) > 255 then
          AFindResult := FindFirst(GetVeryLongFileName(strpas(buf)),faAnyFile,s)
        else
          AFindResult := FindFirst(strpas(buf),faAnyFile,s);
        if AFindResult = 0 then begin
          pData:=pEnd;
          if (s.attr and fadirectory<>0) then begin
            strlcat(buf,'\',sizeof(buf)-1);
            s.size:=0;
          end else if SrcPath[0]<>#0 then begin               {Store only file names!}
            pData:=strrscan(pEnd,'\');
            if pData<>nil then inc(pData)
                          else pData:=pEnd;
          end;
          //Unpacktime(s.time,tdt);
          ConvertTime(s.FindData, tdt);

          Attr := MakeAttrString(s);
          //if Attr <> '' then
            Attr := #9 + Attr;

          ATmp := strpas(pData);

          ATMP2 := AnsiString(DoUTF8Encode(ATmp+#9+FloatToStrF(makecomp(s.size,s.finddata.nFileSizeHigh),ffFixed,18,0)+#9+
            inttostr(tdt.year)+'.'+LeftPad(tdt.month)+'.'+LeftPad(tdt.day)+#9+
            LeftPad(tdt.hour)+':'+LeftPad(tdt.min)+'.'+LeftPad(tdt.sec)+Attr));
          StrLCopy(Atmp1, PAnsiChar(ATMP2), SizeOf(Atmp1) - 1);

          writeln(PackedHandle, Atmp1);

          sysutils.FindClose(s);

          if @PackerProcessDataProc<>nil then
            if PackerProcessDataProc(pwidechar(pEnd),s.size)=0 then
              RunOK:=false;

        end;
        pEnd[0]:=#0;
        p:=strend(p)+1;
      end;
      CloseFile(PackedHandle);
    end;
  end;
  if RunOK then result:=0
           else result:=E_EABORTED;
end;

function PackFiles(PackedFile,SubPath,SrcPath,AddList:pchar;Flags:integer):integer; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function DeleteFilesW(PackedFile, DeleteList: pwidechar): integer; stdcall;
var
  f, tmpfile: textfile;
  tempfn, tmpstr, processedpath, dirtodel, lastdir, ufile: string;
  i, APos: integer;
  A: RawByteString;
  SL: TStringList;
  NeedDeleteFile, NeedDeleteDir, IsDir: boolean;
begin
  Result := E_ECREATE;

  dirtodel := '';
  lastdir := '';
  NeedDeleteDir := false;

  try
    SL := TStringList.Create;
    try
      if (not ParseList(SL, DeleteList)) or (SL.Count = 0) then
        exit;

      if UTF8Only then
        assignfile(f, PackedFile, 65001)
      else
        assignfile(f, PackedFile);
      reset(f);
      try

        i := 0;
        tempfn := PackedFile + '.tmp';
        while FileExists(tempfn) do
        begin
          tempfn := PackedFile + '.tmp' + inttostr(i);
          Inc(i);
        end;

        if UTF8Only then
          assignfile(tmpfile, tempfn, 65001)
        else
          assignfile(tmpfile, tempfn);
        rewrite(tmpfile);
        try
          while not eof(f) do
          begin
            NeedDeleteFile := false;
            IsDir := false;

            if UTF8Only then
            begin
              readln(f, ufile);
              tmpstr := ufile;
            end else begin
              readln(f, A);
              if IsBufferUtf8(pansichar(A), true) then
                tmpstr := UTF8ToUnicodeString(A)
              else
                tmpstr := string(A);
            end;

            APos := AnsiPos(#9, tmpstr);
            if APos > 0 then
              Delete(tmpstr, APos, MaxInt);

            if Length(tmpstr) > 0 then
            begin
              IsDir := tmpstr[Length(tmpstr)] = '\';
              if IsDir then
              begin
                lastdir := tmpstr;
              end;

              for i := 0 to Pred(SL.Count) do
              begin
                if IsDir then
                begin

                  NeedDeleteDir := AnsiSameText(SL[i], tmpstr + '*.*');
                  if not NeedDeleteDir then
                    NeedDeleteDir := (Length(dirtodel) > 0) and
                      (AnsiPos(dirtodel, tmpstr) = 1);
                  if NeedDeleteDir then
                  begin
                    if dirtodel = '' then
                      dirtodel := tmpstr;
                  end
                  else
                    dirtodel := '';
                  NeedDeleteFile := NeedDeleteDir;

                end
                else
                begin

                  NeedDeleteFile := NeedDeleteDir;
                  if not NeedDeleteFile then
                  begin
                    NeedDeleteFile := AnsiSameText(SL[i], lastdir + tmpstr);
                    if (not NeedDeleteFile) and (Pos('\', tmpstr) = 0) then // delete from root
                      NeedDeleteFile := AnsiSameText(SL[i], tmpstr);
                  end;

                end;

                if NeedDeleteFile then
                  break;
              end;
            end;

            if not NeedDeleteFile then
            begin
              if UTF8Only then
                writeln(tmpfile, DoUTF8Encode(ufile))
              else
                writeln(tmpfile, A);
            end;
          end;

          Result := 0;
        finally
          Close(tmpfile);
        end;
      finally
        Close(f);
      end;

      if FileExists(tempfn) then
      begin
        for i := low(ArchiveList) to high(ArchiveList) do
        begin
          if AnsiSameText(ArchiveList[i].ArchiveName, PackedFile) then
          begin
            if ArchiveList[i].ArchiveActive then
            begin
{$I-}
              Close(ArchiveList[i].ArchiveHandle);
{$I+}
              if ioresult <> 0 then
                Result := E_ECLOSE
              else
                Result := 0;
            end;
            ArchiveList[i].ArchiveActive := false;
          end;
        end;

        DeleteFile(PackedFile);
        MoveFile(pwidechar(tempfn), PackedFile);
      end;

    finally
      SL.Free;
    end;
  except
    Result := E_ECREATE;
  end;
end;

function DeleteFiles(PackedFile,DeleteList:pchar):integer; stdcall;
begin
  result:=E_NOT_SUPPORTED;
end;

function GetPackerCaps:integer; stdcall;
begin
  result:=PK_CAPS_NEW or PK_CAPS_MULTIPLE or PK_CAPS_OPTIONS;
end;

procedure ConfigurePacker(ParentHandle,DllInstance: THandle); stdcall;
var
  s: string;
begin
  if UTF8Only then
    s := #13#10 + #13#10 + 'UTF8 Only Mode'
  else
    s := '';

  MessageBox(ParentHandle, PWideChar('DiskdirW plugin, Copyright © 2013-2022 by Progman13' +
    #13#10 + 'Based on sources of Diskdir plugin by Christian Ghisler' + s), 'About diskdirW',mb_iconinformation);
end;

procedure PackSetDefaultParams(dps: Pointer); stdcall;
type
  pPackDefaultParamStruct=^tPackDefaultParamStruct;
  tPackDefaultParamStruct=record
    size,
    PluginInterfaceVersionLow,
    PluginInterfaceVersionHi:longint;
    DefaultIniName:array[0..259] of ansichar;
  end;
var
  LCPInfo: TCPInfo;
begin
  if GetCPInfo(DefaultSystemCodePage, LCPInfo) then
  begin
    UTF8Only := LCPInfo.MaxCharSize > 1;
    //MessageBoxA(GetActiveWindow, (pPackDefaultParamStruct(dps)^.DefaultIniName), nil, 0);
  end;
  with TIniFile.Create(ChangeFileExt(GetModuleName(HInstance), '.ini')) do
  try
    UTF8Only := ReadBool('diskdirW', 'UTF8Only', UTF8Only);
  finally
    Free;
  end;
end;

exports
  OpenArchive,
  OpenArchiveW,
  CloseArchive,
  ReadHeader,
  ReadHeaderExW,
  ProcessFile,
  ProcessFileW,
  SetChangeVolProc,
  SetChangeVolProcW,
  SetProcessDataProc,
  SetProcessDataProcW,
  PackFiles,
  PackFilesW,
  DeleteFiles,
  DeleteFilesW,
  GetPackerCaps,
  ConfigurePacker,
  PackSetDefaultParams;

begin
  FillChar(ArchiveList,sizeof(ArchiveList),#0);
end.
